function [U_cells, infos, options] = parallel_decentralized_MTL_feat(U_cells0, problem, options)
% Parallel implementation of the gossip algorithm.
% min 0.5sum_i || X_iUw_i - y_i || + 0.5*lambda*||w_i||^2  subject to U'U = I.
% 
% options.update = @Grassmann_update and @Grassmann_preconditioned_update.
%                  Types of gradient updates.
% options.lambda = Regularization parameter for ridge regression on W.
% options.stepsize = Initial stepsize.
% options.stepsize_decay = Decay rate parameter.
% options.iters = Number of iterations.
% options.rho = Weight parameter that weighs completion and consensus
%               terms.
% B. Mishra, H. Kasai, P. Jawanpuria, and A. Saroop
% "A Riemannian gossip approach to decentralized subspace learning on Grassmann manifold,"
% Technical report, arXiv preprint arXiv:1705.00467, 2017.


% Original authors: Bamdev Mishra <bamdevm@gmail.com>.
%
        
    % Mandatory problem information
    m = size(problem.data(1).X, 2); % Number of features.
    r = problem.r; % Rank
    N =  problem.N; % Number of agents each with some number of tasks.
    T = length(problem.data); % Total number of tasks.

    % Common computaions
    for t = 1 : T % Number of tasks
        problem.data(t).XtX = problem.data(t).X'*problem.data(t).X;
        problem.data(t).Xty = problem.data(t).X'*problem.data(t).y;
    end
    
    Xtymat = zeros(m,T);
    for t = 1: T % Number of tasks
        Xtymat(:,t) = problem.data(t).Xty;
    end
    
    % Define the Grassmann manifold
    problem.M = grassmannfactory(m, r);
    
    % Distributed the tasks among the agents equally.
    l = floor(T/N);
    Tperm = randperm(T);
    for ii = 1 : N-1
        agent_tasks{ii} = Tperm(1 + (ii-1)*l: min(ii*l, T));
    end
    agent_tasks{N} = Tperm(1 + (N-1)*l: T);
    problem.agent_tasks = agent_tasks;
    
    
    % Initialization of U at all the N nodes
    if isempty(U_cells0)
        for ii = 1:N
            U_cells0{ii} = problem.M.rand();
        end
    end
    U_cells = U_cells0;
    
    % Options
    if ~isfield(options, 'lambda'); options.lambda = 1e-8; end % Regularization parameter.
    if ~isfield(options, 'stepsize'); options.stepsize = 1; end % Initial stepsize.
    if ~isfield(options, 'stepsize_decay'); options.stepsize_decay = 1e3; end % Stepsize decay parameter.
    if ~isfield(options,'iters'); options.iters = 200; end; % 100 updates of nodes 1 and N and 200 updates of the rest.
    if ~isfield(options, 'rho'); options.rho = 1e3; end % Weight parameter
    if ~isfield(options, 'verbosity'); options.verbosity = true; end % Verbose.
    
    stepsize0 = options.stepsize;
    
    % Intialization for collecting the statistics
    for ii = 1 : N
        infos{ii}.cost = [];
        infos{ii}.commcost=[];
        infos{ii}.time = 0;
        infos{ii}.tasks = agent_tasks{ii};
        if ~isempty(problem.data_test) % If there exists a test set.
            infos{ii}.cost_test = [];
        end
    end
    
    % The N+1 structure is used for storing other information.
    infos{N+1}.time = 0;
    infos{N+1}.iter = [];
    
    round1 = 2:2:N;
    round2 = 3:2:N;
    
    % Main algorithm.
    % We assume that the nodes are labeled according to their proximity
    % as follows:
    % for i <= N - 1 the nodes i and i + 1 are neighbors.
    % We pick a node i (<= N-1) uniformly at random.
    %
    % In the parallel mode, round1 denotes picking 1, 3, .. nodes in
    % paralle. In round2, we pick 2, 4, ... in parallel.
    
    
    if exist('parpool', 'file') > 0
        poolobj = gcp('nocreate'); % If no pool, do not create new one.
        
        if isempty(poolobj)
            parpool('local');
        end
        
    else
        if  matlabpool('size') == 0
            matlabpool('open'); % Old matlab
        end
    end
    
    % Stepsize sequence rate parameters
    a = stepsize0;
    b = stepsize0*options.stepsize_decay;
    
    
    for kk = 1 : options.iters % Each iteration picks a set of nodes to update them in parallel.
       
        % Stepsize is decreased at every iteration.
        options.stepsize = a/(1 + b*kk);
        
        % Select the round to update.
        current_round = round1; % Default
        if randi(2) == 2
            % Round 2
            current_round = round2;
        end
        
        % The parallel loop for updating the nodes with a gradient step.
        Uii_current_round = cell(length(current_round), 1);
        Ujj_current_round = cell(length(current_round), 1);
        timeelapsed = Par(length(current_round)); % time storage.
        parfor ll = 1:length(current_round)
            Par.tic; 
            [Uii_current_round{ll}, Ujj_current_round{ll}] = Grassmann_update(current_round(ll)-1, current_round(ll), U_cells, problem, options);
            timeelapsed(ll) = Par.toc; 
        end
        stop(timeelapsed); % Stop par time.
        
        maxtimeelapsed = max([timeelapsed.ItStop] - [timeelapsed.ItStart]); % Time limiting worker.
        
        
        % Update U_cells
        for ll = 1:length(current_round)
            U_cells{current_round(ll) - 1} = Uii_current_round{ll};
            U_cells{current_round(ll)} = Ujj_current_round{ll};
        end
        
        % Collect stats: time
        for ll = 1 : length(current_round)
            ii = current_round(ll) - 1;
            jj = ii + 1;
            infos{ii}.time = [infos{ii}.time; infos{ii}.time(end) + maxtimeelapsed];
            infos{jj}.time = [infos{jj}.time; infos{jj}.time(end) + maxtimeelapsed];
        end
        
        % Collect stats per iteration
        infos{N +1}.time = [infos{N +1}.time; infos{N +1}.time(end) + maxtimeelapsed];
        infos{N +1}.iter = [infos{N +1}.iter; kk];
        
        
        % Collect stats: cost
        W_cells = cell(N,1);
        for ll = 1 : length(current_round)
            ii = current_round(ll) - 1;
            jj = ii + 1;
            
            U_ii = U_cells{ii};
            U_jj = U_cells{jj};
            
            [~, cost1, W_ii] = computegradcost(U_ii, ii, problem, options);
            [~, cost2, W_jj] = computegradcost(U_jj, jj, problem, options);
            
            commcost12 = problem.M.norm(U_ii, problem.M.log(U_ii, U_jj))^2;
            
            
            % Update stats
            infos{ii}.cost = [infos{ii}.cost; cost1];
            infos{ii}.commcost = [infos{ii}.commcost; commcost12];
            infos{jj}.cost = [infos{jj}.cost; cost2];
            
            % Store W_cells
            W_cells{ii} = W_ii;
            W_cells{jj} = W_jj;
        end
        
        
        % Print
        if options.verbosity
            ii = max(current_round(1) - 1, 1);
            jj = ii + 1;
            fprintf('Iteration %0.3d, cost node %d: %7.3e, cost node %d: %7.3e, communication cost: %7.3e, Stepsize %7.3e\n', kk, ii, infos{ii}.cost(end), jj, infos{jj}.cost(end), infos{ii}.commcost(end), options.stepsize);
        end
        
        % Collect stats: test
        if ~isempty(problem.data_test)
            for ll = 1:length(current_round)
                
                ii = current_round(ll) - 1;
                jj = ii + 1;
                
                U_ii = U_cells{ii};
                U_jj = U_cells{jj};
                
                W_ii = W_cells{ii};
                W_jj = W_cells{jj};
                
                cost1_test =  computecosttest(U_ii, ii, problem, options, W_ii);
                cost2_test =  computecosttest(U_jj, jj, problem, options, W_jj);
                
                
                infos{ii}.cost_test = [infos{ii}.cost_test; cost1_test];
                infos{jj}.cost_test = [infos{jj}.cost_test; cost2_test];
                  
            end
            
        end
        
    end
    
    % Make the length of time and cost vectors the same. The difference was
    % due to zero initialization of the time vectors.
    for ii = 1 : N
        infos{ii}.time = infos{ii}.time(2:end);
    end
    infos{N+1}.time = infos{N+1}.time(2:end);
    options.stepsize = stepsize0; % Reset, it was earlier changed.
end


